﻿Imports System.Collections.Generic
Imports System.Linq
Imports System.Data.Common
Imports System.Data.OleDb
Imports System.ComponentModel
Imports System.Reflection
Imports System.Windows.Forms
Imports System.Text
Imports System.Text.RegularExpressions
Imports System.Data.SqlServerCe

Namespace System.Data
#Region "NoCompile"
#If NoCompile Then
hier nur Sub New des TableAdapters: Die Konfiguration ist recht kompliziert, denn er muß für jede Spalte ein DBParameterTemplate speichern, damit andernorts aus sql-Ausdrücken parametrisierte Select-Commands konstruiert wern können.
Entnommen wern die ParamTemplates dem InsertCommand, und für Autowert-ID-Spalten dem Delete-Command, denn für Autowert-IDs generiert der CommandBuilder keine Insert-Parameter.
Die ParamTemplates wern mit eindeutigen ParameterNames ausgerüstet - aber jetzt muß auch der generierte CommandText die neuen ParameterNames addressieren, es sei denn, er addressiert Parameter einfach mit '?'.
Und natürlich alle TableNames und ColumnNames der Command-Texte in '[]' wrappern, evtl. bei MySql in ´´.
Dann wird die Save-Strategie vorbereitet: Access und SqlCe schicken jedem Insert einen Select hinterher, um den von der DB vergebenen Primkey zu erfragen. SqlServer unterstützt UpdateRowSource.Both - man kann also einen Select gleich an den InsertCommandText anhängen, und er aktualisiert dann alle Werte, die u.U. ja vonne DB gesetzt wurden.
Bei Primkey vom Typ Guid wird eingerichtet, dass bei TableNewRow gleich ein Guid generiert wird.
#End If
#End Region 'NoCompile

   Partial Class TableAdapter
      Private Shared _BindFlags As BindingFlags = BindingFlags.Instance Or BindingFlags.NonPublic

      Public Event RowUpdated As EventHandler(Of RowUpdatedEventArgs)

      Public Sub New(ByVal tbl As DataTable, ByVal dtsAttach As DatasetAdapter, ByVal rank As Integer)
         Table = tbl
         _DtsAttach = dtsAttach
         Me.Rank = rank
         ColumnSelector = dtsAttach.Quote(tbl.TableName) & ".* FROM "
         ChildTables = (New DataTable() {tbl}) _
            .Concat(tbl.GetAllChildRelations.Select(Function(rl) rl.ChildTable)).ToArray
         'der Commandbuilder öffnet die Connection des Select-Commands und liest die Tabellendefinition aus. Damit konstruiert er delete/insert/update - Commands. Solange er nicht disposed liefert er dem Adapter alle Commands für jeden Datensatz einzeln. Beim Disposen setzt er im Adapter "seine" delete/insert/updates wieder auf Nothing - deshalb verwendich hier Clone.
         Dim cmdSel = _DtsAttach.CreateSelectCommand(tbl.TableName, False, "", New Object() {})
         Adapter = _DtsAttach.Factory.CreateDataAdapter
         With Adapter
            .SelectCommand = cmdSel
            Dim builder = dtsAttach.CommandBuilder
            builder.DataAdapter = Adapter
            .TableMappings.Add(DbDataAdapter.DefaultSourceTableName, tbl.TableName)
            Try
               .DeleteCommand = builder.GetDeleteCommand.Clone
               .UpdateCommand = builder.GetUpdateCommand.Clone
               .InsertCommand = builder.GetInsertCommand.Clone
               'get a Param for each Column
               For Each p As DbParameter In .InsertCommand.Parameters
                  dtsAttach.AddParam(tbl, p)
               Next
               dtsAttach.AddParam(tbl, .DeleteCommand.Parameters(0))
               AllCommands.AddRange({ _
                  .SelectCommand, .DeleteCommand, .InsertCommand, .UpdateCommand})
            Catch ex As InvalidOperationException
               Console.WriteLine("Für folgenden CommandText schlägt die vollständige Initialisierung des TableAdapters fehl:".And(Environment.NewLine, cmdSel.CommandText))
            End Try
         End With
         Dim PrimKey As DataColumn() = tbl.PrimaryKey
         If PrimKey.Length <> 1 Then Return
         Dim primCol As DataColumn = PrimKey(0)
         If primCol.DataType Is GetType(Guid) Then
            AddHandler primCol.Table.TableNewRow, AddressOf _Table_NewRow
         ElseIf primCol.AutoIncrement Then
            _InitAutoIncrement(primCol)
         End If
      End Sub

      Private Sub _InitAutoIncrement(ByVal primCol As DataColumn)
         Select Case _DtsAttach.Provider
            Case "System.Data.SqlClient"
               'SqlServer eleganter: einfach einen Select ans Insertcommand hängen, und autogenerierte Werte werden zurückgelesen.
               _CreateSmartInsertCommand(primCol, _DtsAttach.Quote, "SCOPE_IDENTITY")
            Case "MySql.Data.MySqlClient"
               _CreateSmartInsertCommand(primCol, _DtsAttach.Quote, "last_insert_id")
            Case Else
               With TryCast(Adapter, OleDbDataAdapter)
                  If .NotNull Then AddHandler .RowUpdated, AddressOf _Adapter_RowUpdated : Return
               End With
               With TryCast(Adapter, SqlCeDataAdapter)
                  If .NotNull Then AddHandler .RowUpdated, AddressOf _Adapter_RowUpdated : Return
               End With
         End Select
      End Sub

      Private Sub _Adapter_RowUpdated(ByVal sender As Object, ByVal e As RowUpdatedEventArgs)
         RaiseEvent RowUpdated(sender, e)
      End Sub

      Private Sub _CreateSmartInsertCommand(ByVal primCol As DataColumn, _
            ByVal quote As Func(Of String, String), ByVal identityFunctionName As String)
         With New ColumnInfo(primCol, Nothing, quote)
            Adapter.InsertCommand.CommandText &= ";".And( _
               "SELECT ", .ColName, " FROM ", .TableName, _
               " WHERE ", .ColName, " = ", identityFunctionName, "()")
         End With
         Adapter.InsertCommand.UpdatedRowSource = UpdateRowSource.Both
      End Sub

      Private Sub _Table_NewRow(ByVal sender As Object, ByVal e As DataTableNewRowEventArgs)
         e.Row(e.Row.Table.PrimaryKey(0)) = Guid.NewGuid
      End Sub

   End Class 'Partial Class TableAdapter

End Namespace
