Imports System.Data.Common
Imports System.Reflection
Imports System.ComponentModel
Imports System.Windows.Forms
Imports System.IO
Imports System.Drawing.Design
Imports System.Text

Namespace System.Data

   ''' <summary>ldt und speichert ein Dataset in Xml auf Platte</summary>
   ''' <remarks>
   ''' Fr "kleine" Einzelplatz-Anwendungen durchaus ausreichend: Bsp.:
   ''' Access-DB: 1,2MB, als Xml-Dataset: 640KB - in 0.2s geladen, in 0.08s gespeichert.
   ''' </remarks>
   Public Class DataSetSaver(Of T As {DataSet, New})

      Public ReadOnly DataSet As T
      ''' <summary>u can replace _Dataset.HasChanges() by a userdefined checking for changes</summary>
      Public Property HasChanges As Func(Of Boolean)
      Public Property DataFile() As String

      Public Event IsLoadingChanged As EventHandler
      Private _IsLoading As Boolean
      Public Property IsLoading() As Boolean
         Get
            Return _IsLoading
         End Get
         Set(ByVal NewValue As Boolean)
            _IsLoading = NewValue
            RaiseEvent IsLoadingChanged(Me, EventArgs.Empty)
         End Set
      End Property

      Public Sub New()
         MyClass.New(New T)
      End Sub

      Public Sub New(ByVal DataSet As T)
         Me.DataSet = DataSet
         HasChanges = AddressOf DataSet.HasChanges
      End Sub

      Public Function CreateForm(Of Tf As {Form, New})() As Tf
         CreateForm = New Tf
         SetMainSource(CreateForm)
      End Function

      Public Sub Load()
         IsLoading = True
         If File.Exists(_DataFile) Then
            'Dim _RankedTables = New Stack(Of DataTable)(DataSet.GetRankedTables)
            With DataSet
               .Clear()
               '_RankedTables.ForEach(Sub(tb) tb.BeginLoadData())
               .LoadCompact(_DataFile)
               '_RankedTables.ForEach(Sub(tb) tb.EndLoadData())
               .AcceptChanges()
            End With
         Else
            MsgDsn("Datei\n", Path.GetFullPath(_DataFile), _
               "\nnicht gefunden.\nVielleicht knnen Sie fortfahren.")
         End If
         IsLoading = False
      End Sub

      Public Sub Save(Optional ByVal checkForChanges As Boolean = True)
         Try
            If checkForChanges AndAlso Not HasChanges.Invoke() Then Return
            With DataSet
               .SaveCompact(_DataFile)
               .AcceptChanges()
            End With
         Finally
            System.Media.SystemSounds.Asterisk.Play()
         End Try
      End Sub

      Public Sub SaveForm(ByVal frm As Form, Optional ByVal checkForChanges As Boolean = True)
         If Not frm.Validate Then Return
         Save(checkForChanges)
      End Sub

      Public ReadOnly HandleFormClosing As FormClosingEventHandler _
         = Sub(sender As Object, e As CancelEventArgs)
              If e.Cancel Then Return
              Dim frm = DirectCast(sender, Form)
              If Not HasChanges.Invoke() Then Return
              Select Case MessageBox.Show(frm, "nderungen speichern?", "Das Ende ist nahe", MessageBoxButtons.YesNoCancel)
                 Case Windows.Forms.DialogResult.Yes
                    Save(False)
                 Case Windows.Forms.DialogResult.Cancel
                    e.Cancel = True
                    Return
              End Select
              e.Cancel = False
           End Sub

      '''' <summary>walks through all members and sets all Datasources of Type T to my Dataset</summary>
      Public Sub SetMainSource(ByVal ctl2 As Control)
         Dim bindFlags As BindingFlags = BindingFlags.Instance Or BindingFlags.Public _
            Or BindingFlags.NonPublic Or BindingFlags.GetField
         Dim shouldReplace As Func(Of Object, Boolean) = Function(src As Object) _
            src IsNot Nothing AndAlso src Is GetType(T) OrElse TypeOf src Is T AndAlso src IsNot DataSet
         Dim recursive As Action(Of Control) = Nothing
         recursive = _
            Sub(ctl)
               'Dbg(ctl)
               'If ctl.Name = "" Then Return
               For Each fld In ctl.GetType.GetFields(bindFlags)
                  'Dim s = fld.FieldType.Name
                  If fld.FieldType Is GetType(T) Then fld.SetValue(ctl, DataSet)
                  If Not GetType(Component).IsAssignableFrom(fld.FieldType) Then Continue For
                  With fld.FieldType.GetProperty("DataSource")
                     If .Null Then Continue For
                     Dim itm = fld.GetValue(ctl)
                     If shouldReplace(.GetValue(itm, Nothing)) Then .SetValue(itm, DataSet, Nothing)
                  End With
               Next
               ctl.Controls.OfType(Of Control).ForEach(recursive)
            End Sub
         recursive(ctl2)
      End Sub

   End Class

End Namespace
