﻿Imports System.ComponentModel
Imports System.Collections.Generic
Imports System.Linq
Imports System.Text

Namespace System

   Public Enum Updates
      GetsUpdated = 1
      Others = 2
      Both = 3
   End Enum

   'damit kann man die Dinger aneinanderhängen, auf Wunsch als Result auch zuweisen. (flow interface)
   Public Interface IPropertyMerge(Of T)
      Function Merge(ByVal other As ActiveProperty(Of T), ByVal mergeMode As Updates) As IPropertyMerge(Of T)
      ReadOnly Property Result() As ActiveProperty(Of T)
   End Interface

   Public Class ActiveProperty(Of T)
      Implements IPropertyMerge(Of T)
      Implements IDisposable

      Public Class EventArg
         Inherits EventArgs
         Public ReadOnly NewValue As T
         Public Sub New(ByVal value As T)
            NewValue = value
         End Sub
      End Class
      Public Custom Event Changed As EventHandler(Of EventArg)
         AddHandler(ByVal value As EventHandler(Of EventArg))
            OnChanged = DirectCast([Delegate].Combine(OnChanged, value), EventHandler(Of EventArg))
         End AddHandler
         RemoveHandler(ByVal value As EventHandler(Of EventArg))
            [Delegate].Remove(OnChanged, value)
         End RemoveHandler
         RaiseEvent(ByVal sender As Object, ByVal e As ActiveProperty(Of T).EventArg)
            OnChanged(sender, e)
         End RaiseEvent
      End Event

      Private OnChanged As EventHandler(Of EventArg) = AddressOf EventDummi
      Private Raise As Action(Of ActiveProperty(Of T), EventArg) = AddressOf EventDummi

      Private _e As New EventArg(Nothing)
      Private Merger As Merging = Nothing
      Private _mergeMode As Updates

      Public Property Value() As T
         Get
            Return _e.NewValue
         End Get
         Set(ByVal value As T)
            Assign(value, True)
         End Set
      End Property
      Public Function AssignOwner(ByVal value As T) As Boolean
         Return Assign(value, False)
      End Function

      Private Function Assign(ByVal value As T, ByVal byForeign As Boolean) As Boolean
         If Object.Equals(_e.NewValue, value) Then Return False
         ResetBindings(value, byForeign)
         Return True
      End Function
      Public Function Assign(ByVal other As ActiveProperty(Of T)) As Boolean
         Return Assign(other._e.NewValue, True)
      End Function
      Private Sub ResetBindings(ByVal value As T, ByVal byForeign As Boolean)
         _e = New EventArg(value)
         If byForeign Then OnChanged(Me, _e)
         If (_mergeMode And Updates.Others) <> 0 Then Raise(Me, _e)
      End Sub
      Public Sub ResetBindings(ByVal value As T)
         ResetBindings(value, True)
      End Sub

      Public Sub SetListener(Of T2)(ByVal obj As Object, ByVal prop As String, ByVal converter As Converter(Of T, T2))
         Dim prop1 = TypeDescriptor.GetProperties(obj.GetType)(prop)
         Dim ValChanged As EventHandler(Of EventArg) = Function(s, e) PropHelper(obj, prop1, converter)
         AddHandler Changed, ValChanged
      End Sub

      Private Function PropHelper(Of T2)(ByVal obj As Object, ByVal prop1 As PropertyDescriptor, ByVal converter As Converter(Of T, T2)) As Object
         prop1.SetValue(obj, converter(_e.NewValue))
         Return Nothing
      End Function

      Public Sub SetListener(ByVal obj As Object, ByVal prop As String)
         Dim tc = TypeDescriptor.GetConverter(obj)
         Dim prop1 = tc.GetProperties(obj)(prop)
         Dim ValChanged As EventHandler(Of EventArg) = Function(s, e) PropHelper(obj, prop1)
         AddHandler Changed, ValChanged
      End Sub

      Private Function PropHelper(ByVal obj As Object, ByVal prop1 As PropertyDescriptor) As Object
         prop1.SetValue(obj, _e.NewValue)
         Return Nothing
      End Function

      Public Function Merge(ByVal myMerge As Updates) As IPropertyMerge(Of T)
         _mergeMode = myMerge
         Return DirectCast(Me, IPropertyMerge(Of T))
      End Function

      Private Function IPropertyMerge_Merge(ByVal other As ActiveProperty(Of T), ByVal mergeMode As Updates) As IPropertyMerge(Of T) Implements IPropertyMerge(Of T).Merge
         other._mergeMode = mergeMode
         If Merger IsNot Nothing Then
            If other.Merger Is Nothing Then
               Merger.Members.Add(other)
               other.Raise = Raise
            Else
               For Each ap In other.Merger.Members
                  ap.Raise = Raise
               Next
               Merger.Members.AddRange(other.Merger.Members)
            End If
            other.Merger = Merger
         ElseIf other.Merger IsNot Nothing Then
            other.Merger.Members.Add(Me)
            Raise = other.Raise
            Merger = other.Merger
         Else
            Merger = New Merging()
            other.Merger = Merger
            Merger.Members.Add(other)
            Merger.Members.Add(Me)
            Raise = AddressOf Merger.DoRaise
            other.Raise = Raise
         End If
         Return DirectCast(Me, IPropertyMerge(Of T))
      End Function
      Private ReadOnly Property IPropertyMerge_Result() As ActiveProperty(Of T) Implements IPropertyMerge(Of T).Result
         Get
            Return Me
         End Get
      End Property

      Private Class Merging
         Public ReadOnly Members As New List(Of ActiveProperty(Of T))()

         Public Sub DoRaise(ByVal sender As ActiveProperty(Of T), ByVal e As EventArg)
            For Each ap In Members
               If Object.ReferenceEquals(ap, sender) Then
                  Continue For
               End If
               If (ap._mergeMode And Updates.GetsUpdated) = 0 Then
                  Continue For
               End If
               If Object.Equals(ap._e.NewValue, e.NewValue) Then
                  Continue For
               End If
               ap._e = e
               ap.OnChanged(ap, e)
            Next
         End Sub
      End Class
      Protected Overridable Sub Dispose(ByVal disposing As Boolean)
         If Not disposing Then Return
         OnChanged = AddressOf EventDummi
         Raise = AddressOf EventDummi
         Merger = Nothing
      End Sub
      Public Sub Dispose() Implements IDisposable.Dispose
         Dispose(True)
      End Sub

      Private Shared Sub EventDummi(ByVal sender As Object, ByVal e As ActiveProperty(Of T).EventArg)
      End Sub
   End Class
End Namespace
